-- --------------------------------------------------------------------
--
--   Title     :  std_logic_1164 multi-value logic system
--   Library   :  This package shall be compiled into a library 
--             :  symbolically named IEEE.
--             :  
--   Developers:  IEEE model standards group (par 1164)
--   Purpose   :  This packages defines a standard for designers
--             :  to use in describing the interconnection data types
--             :  used in vhdl modeling.
--             : 
--   Limitation:  The logic system defined in this package may
--             :  be insufficient for modeling switched transistors,
--             :  since such a requirement is out of the scope of this
--             :  effort. Furthermore, mathematics, primitives,
--             :  timing standards, etc. are considered orthogonal
--             :  issues as it relates to this package and are therefore
--             :  beyond the scope of this effort.
--             :  
--   Note      :  No declarations or definitions shall be included in,
--             :  or excluded from this package. The "package declaration" 
--             :  defines the types, subtypes and declarations of 
--             :  std_logic_1164. The std_logic_1164 package body shall be 
--             :  considered the formal definition of the semantics of 
--             :  this package. Tool developers may choose to implement 
--             :  the package body in the most efficient manner available 
--             :  to them.
--             :
-- --------------------------------------------------------------------
--   modification history :
-- --------------------------------------------------------------------
--   Version: $Revision: 1.1 $
--   Date:    $Date: 2006-06-08 10:43:58-04 $
-- --------------------------------------------------------------------

package std_logic_1164 is

  -------------------------------------------------------------------    
  -- logic state system  (unresolved)
  -------------------------------------------------------------------    
  type STD_ULOGIC is ('U',              -- Uninitialized
                       'X',             -- Forcing  Unknown
                       '0',             -- Forcing  0
                       '1',             -- Forcing  1
                       'Z',             -- High Impedance   
                       'W',             -- Weak     Unknown
                       'L',             -- Weak     0       
                       'H',             -- Weak     1       
                       '-'              -- Don't care
                       );
  -------------------------------------------------------------------    
  -- unconstrained array of std_ulogic for use with the resolution function
  -- and for use in declaring signal arrays of unresolved elements
  -------------------------------------------------------------------    
  type STD_ULOGIC_VECTOR is array (NATURAL range <>) of STD_ULOGIC;

  -------------------------------------------------------------------    
  -- resolution function
  -------------------------------------------------------------------    
  function resolved (s : STD_ULOGIC_VECTOR) return STD_ULOGIC;


  -------------------------------------------------------------------    
  -- logic state system  (resolved)
  -------------------------------------------------------------------    
  subtype STD_LOGIC is resolved STD_ULOGIC;

  -------------------------------------------------------------------    
  -- unconstrained array of resolved std_ulogic for use in declaring
  -- signal arrays of resolved elements
  -------------------------------------------------------------------    
  subtype STD_LOGIC_VECTOR is (resolved) STD_ULOGIC_VECTOR;

  -------------------------------------------------------------------    
  -- common subtypes
  -------------------------------------------------------------------    
  subtype X01 is resolved STD_ULOGIC range 'X' to '1';    -- ('X','0','1') 
  subtype X01Z is resolved STD_ULOGIC range 'X' to 'Z';   -- ('X','0','1','Z') 
  subtype UX01 is resolved STD_ULOGIC range 'U' to '1';   -- ('U','X','0','1') 
  subtype UX01Z is resolved STD_ULOGIC range 'U' to 'Z';  -- ('U','X','0','1','Z') 

  -------------------------------------------------------------------    
  -- local types and constants for use in declaring operations as
  -- instances of uninstantiated subprograms
  -------------------------------------------------------------------    

  type stdlogic_1d is array (STD_ULOGIC) of STD_ULOGIC;
  type stdlogic_table is array(STD_ULOGIC, STD_ULOGIC) of STD_ULOGIC;

  constant and_table  : stdlogic_table;
  constant nand_table : stdlogic_table;
  constant or_table   : stdlogic_table;
  constant nor_table  : stdlogic_table;
  constant xor_table  : stdlogic_table;
  constant xnor_table : stdlogic_table;
  constant not_table  : stdlogic_1d;

  constant cvt_to_x01  : stdlogic_1d;
  constant cvt_to_x01z : stdlogic_1d;
  constant cvt_to_ux01 : stdlogic_1d;

  -------------------------------------------------------------------    
  -- uninstantiated subprograms of which public operations are instances
  -------------------------------------------------------------------    

  function generic_unary_scalar_op
    generic (constant unary_lookup_table : stdlogic_1d)
    parameter (l                         : STD_ULOGIC) return UX01;

  function generic_unary_vector_op
    generic (constant unary_lookup_table : stdlogic_1d)
    parameter (l                         : STD_ULOGIC_VECTOR) return STD_ULOGIC_VECTOR;

  function generic_binary_scalar_scalar_op
    generic (constant binary_lookup_table : stdlogic_table)
    parameter (l                          : STD_ULOGIC; r : STD_ULOGIC) return UX01;

  function generic_binary_vector_vector_op
    generic (constant binary_lookup_table : stdlogic_table)
    parameter (l : STD_ULOGIC_VECTOR;
               r : STD_ULOGIC_VECTOR) return STD_ULOGIC_VECTOR;

  function generic_binary_vector_scalar_op
    generic (constant binary_lookup_table : stdlogic_table)
    parameter (l                          : STD_ULOGIC_VECTOR; r : STD_ULOGIC) return STD_ULOGIC_VECTOR;

  function generic_binary_scalar_vector_op
    generic (constant binary_lookup_table : stdlogic_table)
    parameter (l                          : STD_ULOGIC; r : STD_ULOGIC_VECTOR) return STD_ULOGIC_VECTOR;

  function generic_unary_vector_reduction_op
    generic (constant binary_lookup_table : stdlogic_table;
             constant identity_value      : STD_ULOGIC)
    parameter (l : STD_ULOGIC_VECTOR) return STD_ULOGIC;

  function generic_scalar_conversion_op
    generic (constant conversion_table : stdlogic_1d)
    parameter (s                       : STD_ULOGIC) return STD_ULOGIC;

  function generic_vector_conversion_op
    generic (constant conversion_table : stdlogic_1d)
    parameter (s                       : STD_ULOGIC_VECTOR) return STD_ULOGIC_VECTOR;

  -------------------------------------------------------------------    
  -- overloaded logical operators
  -------------------------------------------------------------------    

  function "and" is new generic_binary_scalar_scalar_op generic map (and_table);
  function "nand" is new generic_binary_scalar_scalar_op generic map (nand_table);
  function "or" is new generic_binary_scalar_scalar_op generic map (or_table);
  function "nor" is new generic_binary_scalar_scalar_op generic map (nor_table);
  function "xor" is new generic_binary_scalar_scalar_op generic map (xor_table);
  function "xnor" is new generic_binary_scalar_scalar_op generic map (xnor_table);
  function "not" is new generic_unary_scalar_op generic map (not_table);

  -------------------------------------------------------------------    
  -- vectorized overloaded logical operators
  -------------------------------------------------------------------    
  function "and" is new generic_binary_vector_vector_op generic map (and_table);
  function "nand" is new generic_binary_vector_vector_op generic map (nand_table);
  function "or" is new generic_binary_vector_vector_op generic map (or_table);
  function "nor" is new generic_binary_vector_vector_op generic map (nor_table);
  function "xor" is new generic_binary_vector_vector_op generic map (xor_table);
  function "xnor" is new generic_binary_vector_vector_op generic map (xnor_table);
  function "not" is new generic_unary_vector_op generic map (not_table);

  function "and" is new generic_binary_vector_scalar_op generic map (and_table);
  function "and" is new generic_binary_scalar_vector_op generic map (and_table);

  function "nand" is new generic_binary_vector_scalar_op generic map (nand_table);
  function "nand" is new generic_binary_scalar_vector_op generic map (nand_table);

  function "or" is new generic_binary_vector_scalar_op generic map (or_table);
  function "or" is new generic_binary_scalar_vector_op generic map (or_table);

  function "nor" is new generic_binary_vector_scalar_op generic map (nor_table);
  function "nor" is new generic_binary_scalar_vector_op generic map (nor_table);

  function "xor" is new generic_binary_vector_scalar_op generic map (xor_table);
  function "xor" is new generic_binary_scalar_vector_op generic map (xor_table);

  function "xnor" is new generic_binary_vector_scalar_op generic map (xnor_table);
  function "xnor" is new generic_binary_scalar_vector_op generic map (xnor_table);

  function "and" is new generic_unary_vector_reduction_op
    generic map (and_table, '1');
  function "nand" is new generic_unary_vector_reduction_op
    generic map (nand_table, '1');
  function "or" is new generic_unary_vector_reduction_op
    generic map (or_table, '0');
  function "nor" is new generic_unary_vector_reduction_op
    generic map (nor_table, '0');
  function "xor" is new generic_unary_vector_reduction_op
    generic map (xor_table, '0');
  function "xnor" is new generic_unary_vector_reduction_op
    generic map (xnor_table, '0');

  -------------------------------------------------------------------
  -- shift operators
  -------------------------------------------------------------------

  function "sll" (l : STD_ULOGIC_VECTOR, r : INTEGER) return STD_ULOGIC_VECTOR;
  function "srl" (l : STD_ULOGIC_VECTOR, r : INTEGER) return STD_ULOGIC_VECTOR;
  function "rol" (l : STD_ULOGIC_VECTOR, r : INTEGER) return STD_ULOGIC_VECTOR;
  function "ror" (l : STD_ULOGIC_VECTOR, r : INTEGER) return STD_ULOGIC_VECTOR;

  -------------------------------------------------------------------
  -- conversion functions
  -------------------------------------------------------------------
  function To_bit (s       : STD_ULOGIC; xmap : BIT        := '0') return BIT;
  function To_bitvector (s : STD_ULOGIC_VECTOR; xmap : BIT := '0') return BIT_VECTOR;

  function To_StdULogic (b       : BIT) return STD_ULOGIC;
  function To_StdLogicVector (b  : BIT_VECTOR) return STD_LOGIC_VECTOR;
  function To_StdLogicVector (s  : STD_ULOGIC_VECTOR) return STD_LOGIC_VECTOR;
  function To_StdULogicVector (b : BIT_VECTOR) return STD_ULOGIC_VECTOR;
  function To_StdULogicVector (s : STD_LOGIC_VECTOR) return STD_ULOGIC_VECTOR;

  alias To_Bit_Vector is
    To_bitvector[STD_ULOGIC_VECTOR return BIT_VECTOR];
  alias To_BV is
    To_bitvector[STD_ULOGIC_VECTOR return BIT_VECTOR];
  
  alias To_Std_Logic_Vector is
    To_StdLogicVector[BIT_VECTOR return STD_LOGIC_VECTOR];
  alias To_SLV is
    To_StdLogicVector[BIT_VECTOR return STD_LOGIC_VECTOR];

  alias To_Std_Logic_Vector is
    To_StdLogicVector[STD_ULOGIC_VECTOR return STD_LOGIC_VECTOR];
  alias To_SLV is
    To_StdLogicVector[STD_ULOGIC_VECTOR return STD_LOGIC_VECTOR];

  alias To_Std_ULogic_Vector is
    To_StdULogicVector[BIT_VECTOR return STD_ULOGIC_VECTOR];
  alias To_SUV is
    To_StdULogicVector[BIT_VECTOR return STD_ULOGIC_VECTOR];

  alias To_Std_ULogic_Vector is
    To_StdULogicVector[STD_LOGIC_VECTOR return STD_ULOGIC_VECTOR];
  alias To_SUV is
    To_StdULogicVector[STD_LOGIC_VECTOR return STD_ULOGIC_VECTOR];

  -------------------------------------------------------------------    
  -- strength strippers and type convertors
  -------------------------------------------------------------------    

  function To_X01 is new generic_vector_conversion_op generic map (cvt_to_x01);
  function To_X01 is new generic_scalar_conversion_op generic map (cvt_to_x01);
  alias To_X01 is To_StdULogicVector [BIT_VECTOR return STD_ULOGIC_VECTOR];
  alias To_X01 is To_StdULogic [BIT return STD_ULOGIC];

  function To_X01Z is new generic_vector_conversion_op generic map (cvt_to_x01z);
  function To_X01Z is new generic_scalar_conversion_op generic map (cvt_to_x01z);
  alias To_X01Z is To_StdULogicVector [BIT_VECTOR return STD_ULOGIC_VECTOR];
  alias To_X01Z is To_StdULogic [BIT return STD_ULOGIC];

  function To_UX01 is new generic_vector_conversion_op generic map (cvt_to_ux01);
  function To_UX01 is new generic_scalar_conversion_op generic map (cvt_to_ux01);
  alias To_UX01 is To_StdULogicVector [BIT_VECTOR return STD_ULOGIC_VECTOR];
  alias To_UX01 is To_StdULogic [BIT return STD_ULOGIC];

  function "??" (l : STD_ULOGIC) return BOOLEAN;

  -------------------------------------------------------------------    
  -- edge detection
  -------------------------------------------------------------------    
  function rising_edge (signal s  : STD_ULOGIC) return BOOLEAN;
  function falling_edge (signal s : STD_ULOGIC) return BOOLEAN;

  -------------------------------------------------------------------    
  -- object contains an unknown
  -------------------------------------------------------------------    
  function Is_X (s : STD_ULOGIC_VECTOR) return BOOLEAN;
  function Is_X (s : STD_ULOGIC) return BOOLEAN;

  -------------------------------------------------------------------
  -- matching relational operators
  -------------------------------------------------------------------
  -- the following operations are predefined

  -- FUNCTION "?="  ( l, r : std_ulogic        ) RETURN std_ulogic;
  -- FUNCTION "?="  ( l, r : std_ulogic_vector ) RETURN std_ulogic;

  -- FUNCTION "?/=" ( l, r : std_ulogic        ) RETURN std_ulogic;
  -- FUNCTION "?/=" ( l, r : std_ulogic_vector ) RETURN std_ulogic;

  -- FUNCTION "?<"  ( l, r : std_ulogic        ) RETURN std_ulogic;
  -- FUNCTION "?<=" ( l, r : std_ulogic        ) RETURN std_ulogic;
  -- FUNCTION "?>"  ( l, r : std_ulogic        ) RETURN std_ulogic;
  -- FUNCTION "?>=" ( l, r : std_ulogic        ) RETURN std_ulogic;

  -------------------------------------------------------------------
  -- string conversion and write operations
  -------------------------------------------------------------------
  -- the following operations are predefined

  -- FUNCTION to_string ( value : std_ulogic ) RETURN string;
  -- FUNCTION to_string ( value : std_ulogic_vector ) RETURN string;

  -- explicitly defined operations

  alias TO_BSTRING is TO_STRING [STD_ULOGIC_VECTOR return STRING];
  alias TO_BINARY_STRING is TO_STRING [BIT_VECTOR return STRING];
  function TO_OSTRING (VALUE : STD_ULOGIC_VECTOR) return STRING;
  alias TO_OCTAL_STRING is TO_OSTRING [BIT_VECTOR return STRING];
  function TO_HSTRING (VALUE : STD_ULOGIC_VECTOR) return STRING;
  alias TO_HEX_STRING is TO_HSTRING [BIT_VECTOR return STRING];

  procedure READ (L : inout LINE; VALUE : out STD_ULOGIC; GOOD : out BOOLEAN);
  procedure READ (L : inout LINE; VALUE : out STD_ULOGIC);

  procedure READ (L : inout LINE; VALUE : out STD_ULOGIC_VECTOR; GOOD : out BOOLEAN);
  procedure READ (L : inout LINE; VALUE : out STD_ULOGIC_VECTOR);

  procedure WRITE (L         : inout LINE; VALUE : in STD_ULOGIC;
                   JUSTIFIED : in    SIDE := right; FIELD : in WIDTH := 0);

  procedure WRITE (L         : inout LINE; VALUE : in STD_ULOGIC_VECTOR;
                   JUSTIFIED : in    SIDE := right; FIELD : in WIDTH := 0);

  alias BREAD is READ [LINE, STD_ULOGIC_VECTOR, BOOLEAN];
  alias BREAD is READ [LINE, STD_ULOGIC_VECTOR];
  alias BINARY_READ is READ [LINE, STD_ULOGIC_VECTOR, BOOLEAN];
  alias BINARY_READ is READ [LINE, STD_ULOGIC_VECTOR];

  procedure OREAD (L : inout LINE; VALUE : out STD_ULOGIC_VECTOR; GOOD : out BOOLEAN);
  procedure OREAD (L : inout LINE; VALUE : out STD_ULOGIC_VECTOR);
  alias OCTAL_READ is OREAD [LINE, STD_ULOGIC_VECTOR, BOOLEAN];
  alias OCTAL_READ is OREAD [LINE, STD_ULOGIC_VECTOR];

  procedure HREAD (L : inout LINE; VALUE : out STD_ULOGIC_VECTOR; GOOD : out BOOLEAN);
  procedure HREAD (L : inout LINE; VALUE : out STD_ULOGIC_VECTOR);
  alias HEX_READ is HREAD [LINE, STD_ULOGIC_VECTOR, BOOLEAN];
  alias HEX_READ is HREAD [LINE, STD_ULOGIC_VECTOR];

  alias BWRITE is WRITE [LINE, STD_ULOGIC_VECTOR, SIDE, WIDTH];
  alias BINARY_WRITE is WRITE [LINE, STD_ULOGIC_VECTOR, SIDE, WIDTH];
  
  procedure OWRITE (L         : inout LINE; VALUE : in STD_ULOGIC_VECTOR;
                    JUSTIFIED : in    SIDE := right; FIELD : in WIDTH := 0);
  alias OCTAL_WRITE is OWRITE [LINE, STD_ULOGIC_VECTOR, SIDE, WIDTH];
  
  procedure HWRITE (L         : inout LINE; VALUE : in STD_ULOGIC_VECTOR;
                    JUSTIFIED : in    SIDE := right; FIELD : in WIDTH := 0);
  alias HEX_WRITE is HWRITE [LINE, STD_ULOGIC_VECTOR, SIDE, WIDTH];
  
end package std_logic_1164;
